#!/usr/bin/env python

import json
import os
import sys
import textwrap

import click
from keyme import KeyMe
from keyme import writer


class Config(writer.ConfigParser):
    def __init__(self, path='~/.aws/keyme', *args, **kwargs):
        super(Config, self).__init__(*args, **kwargs)
        self._path = path

    def load(self):
        path = os.path.expanduser(self._path)
        if os.path.exists(path):
            self.read(path)

    def save(self):
        path = os.path.expanduser(self._path)
        with open(path, 'w') as fp:
            self.write(fp)


def generate_keys(event, context):
    username = event.get('username')
    password = event.get('password')
    mfa_code = event.get('mfa_code')
    region = event.get('region', 'us-east-1')

    # The following, we expect, to come from $stageVariables
    idp = event.get('idpid')
    sp  = event.get('spid')
    role = event.get('role')
    principal = event.get('principal')

    # Duplication is to avoid defaulting values in the class
    # - thats logic we shouldn't be doing there
    return KeyMe(
        username=username,
        password=password,
        mfa_code=mfa_code,
        idp=idp,
        sp=sp,
        region=region,
        role=role,
        principal=principal
    ).key()


pass_config = click.make_pass_decorator(Config, ensure=True)


@click.group(chain=True)
@pass_config
def cli(config):
    config.load()
    pass


@cli.command('show-config')
@pass_config
def show_config(config):
    data = json.dumps(config)
    click.echo(data)


@cli.command('init')
@pass_config
@click.option('--update', help='update configuration for given env name')
def setup(config, update):
    """
    Creates a KeyMe environment populated with the settings necessary for
    subsequent login actions with that environment.

    :param config:
        The configuration parser where the environment data will be stored.
    :param update:
        The name of the environment that will be populated. If omitted, the
        "default" environment is used.
    """

    if update not in config:
        name = click.prompt(
            'Please enter a name for this config',
            default='default'
        )
    else:
        name = update

    idp_id = click.prompt('Please enter your google idp id')
    sp_id = click.prompt('Please enter your aws sp id')
    aws_region = click.prompt(
        'Which AWS Region do you want to be default?',
        default='us-east-1'
    )
    principal_arn = click.prompt('Please provide your default principal ARN')
    role_arn = click.prompt('Please provide your default role arn')
    duration_seconds = click.prompt(
        'Please provide the duration in seconds of the sts token',
        default=3600,
        type=int
    )

    use_mfa = click.confirm('Enable MFA tokens?')
    use_s3v4 = click.confirm('Enable S3 signature version 4?')

    config[name] = dict(
        idpid=idp_id,
        spid=sp_id,
        region=aws_region,
        principal=principal_arn,
        role=role_arn,
        duration_seconds=duration_seconds,
        mfa='yes' if use_mfa else 'no',
        s3v4='yes' if use_s3v4 else 'no'
    )
    config.save()


@cli.command('login')
@pass_config
@click.option(
    '--mfa', '-m',
    is_flag=True,
    help='Enables MFA if disabled in default configuration'
)
@click.option(
    '--s3v4',
    is_flag=True,
    help='Enables S3 signature version 4. Only applicable with the save flag'
)
@click.option('--username', prompt='Please Enter your user name')
@click.option(
    '--password',
    prompt='Please enter your password',
    hide_input=True
)
@click.option('--idp', '-i', help='Allows overrideing of the IDP id')
@click.option('--sp', '-s', help='Allows overriding of the store SP id ')
@click.option(
    '--principal','-a',
    help='Allows overriding of the store principal'
)
@click.option('--role', '-r', help='Allows overriding of the stored role ARN')
@click.option(
    '--region',
    help='Allows changing the aws region by overriding default stored value'
)
@click.option('--env', '-e', help='Environment name given during setup')
@click.option(
    '--duration', '-d',
    help='override stored duration for creds from sts'
)
@click.option(
    '--save',
    is_flag=True,
    help='Write the credentials to the standard AWS configuration files'
)
def login(
        config,
        username,
        password,
        mfa,
        duration,
        idp,
        sp,
        principal,
        role,
        region,
        env,
        save,
        s3v4
):

    try:
        data = config[env or 'default']
    except KeyError:
        click.echo(textwrap.dedent(
            """
            Login failed. The "{}" KeyMe profile is not 
            initialized. Please run the fetch_creds init command 
            first to create this profile or specify a different 
            profile with the --env flag.
            """.format(env or 'default')
        ))
        sys.exit(1)

    if mfa or data.getboolean('mfa'):
        mfa = click.prompt('Please enter MFA Token')
    else:
        mfa = None

    click.echo('Loading "{}" config'.format(env or 'default'))

    payload = {
        'username': username,
        'password': password,
        'mfa_code': mfa,
        'role': role or data['role'],
        'principal': principal or data['principal'],
        'idpid': idp or data['idpid'],
        'spid': sp or data['spid'],
        'region': region or data['region'],
        'duration': duration or data['duration_seconds']
    }
    k = generate_keys(payload, {})
    aws = k['aws']

    if save:
        writer.write_credentials_file(env, **aws)
        writer.write_config_file(
            profile_name=env,
            default_region=region,
            enable_s3v4=s3v4 or data.getboolean('s3v4')
        )

    click.echo('export AWS_ACCESS_KEY_ID=\'%(access_key)s\'' % aws)
    click.echo('export AWS_SECRET_ACCESS_KEY=\'%(secret_key)s\'' % aws)
    click.echo('export AWS_SESSION_TOKEN=\'%(session_token)s\'' % aws)



import os
import shutil

try:
    from configparser import ConfigParser
except ImportError:
    from ConfigParser import ConfigParser

DEFAULT_REGION = 'us-east-1'
DEFAULT_OUTPUT_TYPE = 'json'


def create_backup(source_path):
    """
    Creates a backup copy of the specified file in the same location but with
    a backup extension.
    """

    backup_path = '{}.keyme.backup'.format(source_path)
    try:
        if os.path.exists(backup_path):
            os.remove(backup_path)
        shutil.copy2(source_path, backup_path)
        return True
    except Exception:
        return False


def write_credentials_file(profile_name, access_key, secret_key, session_token):
    """
    Writes the specified AWS credentials to the standard aws credentials
    file in the user's home directory.

    :param profile_name:
        The name of the profile for which the credentials will be written.
    :param access_key:
        The "aws_access_key_id" value for the specified profile.
    :param secret_key:
        The "aws_secret_access_key" value for the specified profile.
    :param session_token:
        The "aws_session_token" value for the specified profile.
    :return:
        A dictionary representation of the contents of the config file after
        it has been modified by this function.
    """

    credentials_path = os.path.expanduser('~/.aws/credentials')
    credentials = ConfigParser()

    if os.path.exists(credentials_path):
        credentials.read(credentials_path)

    credentials[profile_name] = dict(
        aws_access_key_id=access_key,
        aws_secret_access_key=secret_key,
        aws_session_token=session_token
    )

    create_backup(credentials_path)
    with open(credentials_path, 'w') as fp:
        credentials.write(fp)

    return dict(credentials.items())


def write_config_file(
        profile_name,
        default_region=DEFAULT_REGION,
        output_type=DEFAULT_OUTPUT_TYPE,
        enable_s3v4=False
):
    """
    Write an AWS config file for the specified profile, which includes
    information about region, output type and other session configuration
    data.

    :param profile_name:
        The name of the profile for which the config settings will be written.
    :param default_region:
        The default AWS region for the profile.
    :param output_type:
        The default data type returned by console commands for this profile.
    :param enable_s3v4:
        Whether or not to enable signature version 4 when interacting with
        S3 data encrypted by KMS.
    :return:
        A dictionary representation of the contents of the config file after
        it has been modified by this function.
    """

    configs_path = os.path.expanduser('~/.aws/config')
    configs = ConfigParser()

    if os.path.exists(configs_path):
        configs.read(configs_path)

    profile_key = 'profile {}'.format(profile_name)
    if profile_key not in configs:
        configs[profile_key] = {}
    config_data = configs[profile_key]
    config_data['region'] = default_region or DEFAULT_REGION
    config_data['output'] = output_type or DEFAULT_OUTPUT_TYPE

    if enable_s3v4:
        config_data['s3'] = '\nsignature_version = s3v4'
    elif 's3' in config_data:
        del config_data['s3']

    create_backup(configs_path)
    with open(configs_path, 'w') as fp:
        configs.write(fp)

    return dict(configs.items())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function

import sys

from six.moves.urllib.parse import urlencode
import boto3
import requests
from bs4 import BeautifulSoup


class KeyMe:
    def __init__(self, **kwargs):
        """Initialise an Alky object

        Keyword Arguments:
        username -- username **with fqdn** to login with
        password -- corresponding password
        """
        self.username = kwargs.pop('username')
        self.password = kwargs.pop('password')
        self.mfa_code = kwargs.pop('mfa_code')
        self.role = kwargs.pop('role')
        self.principal = kwargs.pop('principal')
        self.session = None
        self.sts = None

        self.region = kwargs.pop('region')
        self.idp = kwargs.pop('idp')
        self.sp = kwargs.pop('sp')
        if kwargs.get('duration_seconds'):
            self.timeout = kwargs.pop('duration_seconds')
        else:
            self.duration_seconds = 3600

        self.google_accounts_url = self.idp_entry_url
        if kwargs:
            raise ValueError("Extraneous keys passed: %s" % kwargs.keys())

    @property
    def idp_entry_url(self):
        """Returns the SAML SSO init URL populated with IDP & SP parameters"""
        url = 'https://accounts.google.com/o/saml2/initsso'
        url_params = dict(
            idpid=self.idp or '',
            spid=self.sp or '',
            forceauthn='false'
        )
        return '{}?{}'.format(url, urlencode(url_params))

    def key(self):
        """Return a key object"""
        self.session = self.login_to_google()
        self.sts = self.login_to_sts(self.region)

        saml = self.parse_google_saml()

        (
            access_key,
            secret_key,
            session_token,
            expiration
        ) = self.get_tokens(saml, self.role, self.principal)

        return {
            'aws': {
                'access_key': access_key,
                'secret_key': secret_key,
                'session_token': session_token
            },
            'expiration': expiration
        }

#    def login_to_google(self, idpid, spid, email, password, mfapin):
    def login_to_google(self):
        # Initiate session handler
        session = requests.Session()

        # Configure Session Headers
        session.headers['User-Agent'] = 'AWS Sign-in'

        # Initial Page load
        google_session = session.get(self.idp_entry_url)
        google_session.raise_for_status()
        session.headers['Referrer'] = google_session.url

        # Collect information from the page source
        decoded = BeautifulSoup(google_session.text, 'html.parser')
        gxf = decoded.find('input', {'name': 'gxf'}).get('value')
        cont = decoded.find('input', {'name': 'continue'}).get('value')
        page = decoded.find('input', {'name': 'Page'}).get('value')
        sign_in = decoded.find('input', {'name': 'signIn'}).get('value')
        account_login_url = (
            decoded
            .find('form', {'id': 'gaia_loginform'})
            .get('action')
        )

        # Setup the payload
        payload = {
            'Page': page,
            'gxf': gxf,
            'continue': cont,
            'ltmpl': 'popup',
            'scc': 1,
            'sarp': 1,
            'oauth': 1,
            'ProfileInformation': '',
            'SessionState': '',
            '_utf8': '?',
            'bgresponse': 'js_disabled',
            'pstMsg': 0,
            'checkConnection': '',
            'checkedDomains': 'youtube',
            'Email': self.username,
            'identifiertoken': '',
            'identifiertoken_audio': '',
            'identifier-captcha-input': '',
            'signIn': sign_in,
            'Passwd': '',
            'PersistentCookie': 'yes',
        }

        # POST to account login info page, to collect profile and session info
        google_session = session.post(account_login_url, data=payload)
        google_session.raise_for_status()
        session.headers['Referrer'] = google_session.url

        # Collect ProfileInformation, SessionState, signIn, and Password
        # Challenge URL
        decoded = BeautifulSoup(google_session.text, 'html.parser')

        profile_information = (
            decoded
            .find('input', {'name': 'ProfileInformation'})
            .get('value')
        )
        session_state = (
            decoded
            .find('input', {'name': 'SessionState'})
            .get('value')
        )
        sign_in = decoded.find('input', {'name': 'signIn'}).get('value')
        passwd_challenge_url = (
            decoded
            .find('form', {'id': 'gaia_loginform'})
            .get('action')
        )

        # Update the payload
        payload['SessionState'] = session_state
        payload['ProfileInformation'] = profile_information
        payload['signIn'] = sign_in
        payload['Passwd'] = self.password

        # POST to Authenticate Password
        google_session = session.post(passwd_challenge_url, data=payload)
        google_session.raise_for_status()
        decoded = BeautifulSoup(google_session.text, 'html.parser')
        error = decoded.find(class_='error-msg')
        cap = decoded.find('input', {'name': 'logincaptcha'})
        if error is not None or cap is not None:
            try:
                raise ValueError(
                    'Wrong Password or Captcha Required. '
                    'Manually Login to remove this.'
                )
            except ValueError as e:
                print(e)
                sys.exit(1)

        session.headers['Referrer'] = google_session.url
        if self.mfa_code is not None:
            # Collect the TL, and Updated gxf
            decoded = BeautifulSoup(google_session.text, 'html.parser')
            tl = decoded.find('input', {'name': 'TL'}).get('value')
            gxf = decoded.find('input', {'name': 'gxf'}).get('value')

            # Dynamically configure TOTP URL and ID based upon the session url
            challenge_url = google_session.url.split("?")[0]
            challenge_id = challenge_url.split("totp/")[1]

            # Create a new payload
            payload = {
                'challengeId': challenge_id,
                'challengeType': 6,
                'continue': cont,
                'scc': 1,
                'sarp': 1,
                'checkedDomains': 'youtube',
                'pstMsg': 0,
                'TL': tl,
                'gxf': gxf,
                'Pin': self.mfa_code,
                'TrustDevice': 'on',
            }
            # Submit TOTP
            google_session = session.post(challenge_url, data=payload)
            google_session.raise_for_status()

        return google_session

    def parse_google_saml(self):
        """Load and parse saml from google"""
        parsed = BeautifulSoup(self.session.text, 'html.parser')
        saml_element = (
            parsed
            .find('input', {'name':'SAMLResponse'})
            .get('value')
        )

        try:
            if not saml_element:
                raise Exception(
                    'Could not get a SAML reponse, check credentials'
                )
        except Exception as e:
            print(e)
            sys.exit(1)

        return saml_element

    @staticmethod
    def login_to_sts(region):
        """Create an STS context via STS"""
        return boto3.client('sts',region_name=region)

    def get_tokens(self, saml, role, principal):
        """Load and parse tokes from AWS STS"""
        token = self.sts.assume_role_with_saml(
            RoleArn=role,
            PrincipalArn=principal,
            SAMLAssertion=saml,
            DurationSeconds=self.duration_seconds
        )
        credentials = token['Credentials']

        return (
            credentials['AccessKeyId'],
            credentials['SecretAccessKey'],
            credentials['SessionToken'],
            credentials['Expiration']
        )



import unittest

from mock import patch
from keyme import writer


class TestWriter(unittest.TestCase):
    """Test suite for the writer module"""

    @patch('builtins.open')
    @patch('os.path.exists')
    @patch('keyme.writer.create_backup')
    @patch('keyme.writer.ConfigParser.write')
    @patch('keyme.writer.ConfigParser.read')
    def test_write_credentials_file(
            self,
            read,
            write,
            create_backup,
            exists,
            builtins_open
    ):
        """Should write a credentials file with the specified info"""

        exists.return_value = True

        result = writer.write_credentials_file(
            profile_name='TEST',
            access_key='ACCESS_KEY',
            secret_key='SECRET_KEY',
            session_token='SESSION_TOKEN'
        )

        self.assertEqual(1, read.call_count)
        self.assertEqual(1, write.call_count)
        self.assertEqual(1, create_backup.call_count)
        self.assertIn('TEST', result)

    @patch('builtins.open')
    @patch('os.path.exists')
    @patch('keyme.writer.create_backup')
    @patch('keyme.writer.ConfigParser.write')
    @patch('keyme.writer.ConfigParser.read')
    def test_write_config_file(
            self,
            read,
            write,
            create_backup,
            exists,
            builtins_open
    ):
        """Should write a config file with the specified info"""

        exists.return_value = True

        result = writer.write_config_file(profile_name='TEST')

        self.assertEqual(1, read.call_count)
        self.assertEqual(1, write.call_count)
        self.assertEqual(1, create_backup.call_count)
        self.assertIn('profile TEST', result)

    @patch('shutil.copy2')
    @patch('os.remove')
    @patch('os.path.exists')
    def test_create_backup(self, exists, remove, copy):
        """Should create a backup copy of the specified file"""

        exists.return_value = True

        path = 'fake'
        backup_path = 'fake.keyme.backup'
        result = writer.create_backup(path)
        self.assertTrue(result)
        self.assertEqual(1, remove.call_count)
        self.assertEqual(remove.call_args[0][0], backup_path)
        self.assertEqual(1, copy.call_count)
        self.assertEqual(copy.call_args[0][0], path)
        self.assertEqual(copy.call_args[0][1], backup_path)

    @patch('os.path.exists')
    def test_create_backup_error(self, exists):
        """Should fail to create a backup copy of the specified file"""

        exists.side_effect = IOError('FAKE')

        result = writer.create_backup('fake-path')
        self.assertFalse(result)



"""
keyme and fetch_creds -- Tools for interacting with GOOGLE SAML SSO and AWS
SAML with STS.
"""

from setuptools import find_packages, setup

dependencies = [
    'click',
    'boto3',
    'beautifulsoup4',
    'requests',
    'py',
    'six'
]

setup(name='keyme',
      version='0.8.0',
      description='Google SAML STS login library',
      long_description=(
          'This tool allows you to login into AWS using Google SAML Apps as '
          'part of Google For Work'
      ),
      url='http://github.com/wheniwork/keyme',
      author='Richard Genthner',
      author_email='richard.genthner@wheniwork.com',
      packages=find_packages(),
      include_package_data=True,
      install_requires=dependencies,
      tests_require=['pytest', 'mock'],
      license='MIT',
      entry_points = '''
        [console_scripts]
            fetch_creds=keyme.cli.fetch_creds:cli
        ''',
      platforms='any',
      zip_safe=False,
      classifiers=[
        # As from http://pypi.python.org/pypi?%3Aaction=list_classifiers
        # 'Development Status :: 1 - Planning',
        # 'Development Status :: 2 - Pre-Alpha',
        # 'Development Status :: 3 - Alpha',
        'Development Status :: 4 - Beta',
        # 'Development Status :: 5 - Production/Stable',
        # 'Development Status :: 6 - Mature',
        # 'Development Status :: 7 - Inactive',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX',
        'Operating System :: MacOS',
        'Operating System :: Unix',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ])
